#include <iostream>
#include <Carbon/Carbon.h>
#include "opn_interface.h"

typedef struct
{
	char*		symbologyName;
	UInt8		symbol;
} SBCD_DEFS;

enum symbology
{
	Bookland = 0x16,
	Codabar = 0x02,
	Code_11 = 0x0c,
	Code_32 = 0x20,
	Code_128 = 0x03,
	Code_39 = 0x01,
	Code_39_Full_ASCII = 0x13,
	Code_93 = 0x07,
	Composite = 0x1d,
	Coupon = 0x17,
	D25 = 0x04,
	Data_Matrix = 0x1b,
	EAN_128 = 0x0f,
	EAN_13 = 0x0b,
	EAN_13_2 = 0x4b,
	EAN_13_5 = 0x8b,
	EAN_8 = 0x0a,
	EAN_8_2 = 0x4a,
	EAN_8_5 = 0x8a,
	IATA = 0x05,
	ISBT_128 = 0x19,
	ISBT_128_concatenated = 0x21,
	ITF = 0x06,
	Macro_PDF = 0x28,
	MSI = 0x0E,
	PDF_417 = 0x11,
	Postbar_Canada = 0x26,
	Postnet_US = 0x1e,
	Postal_Australia = 0x23,
	Postal_Japan = 0x22,
	Postal_UK = 0x27,
	QR_code = 0x1c,
	RSS_limited = 0x31,
	RSS_14 = 0x30,
	RSS_Expanded = 0x32,
	Signature = 0x24,
	Trioptic_Code_39 = 0x15,
	UPCA = 0x08,
	UPCA_2 = 0x48,
	UPCA_5 = 0x88,
	UPCE = 0x09,
	UPCE_2 = 0x49,
	UPCE_5 = 0x89,
	UPCE1 = 0x10,
	UPCE1_2 = 0x50,
	UPCE1_5 = 0x90
};



const SBCD_DEFS bcodedefs[] =
{
	"Bookland",					Bookland,
	"Codabar",					Codabar,
	"Code-11",					Code_11,
	"Code-32",					Code_32,
	"Code-128",					Code_128,
	"Code-39",					Code_39,
	"Code-39 full ASCII",		Code_39_Full_ASCII,
	"Code-93",					Code_93,
	"Composite",				Composite,
	"Coupon",					Coupon,
	"D 2 of 5",					D25,
	"Data Matrix",				Data_Matrix,
	"EAN-128",					EAN_128,
	"EAN-13",					EAN_13,
	"EAN-13 + 2",				EAN_13_2,
	"EAN-13 + 5",				EAN_13_5,
	"EAN-8",					EAN_8,
	"EAN-8 + 2",				EAN_8_2,
	"EAN-8 + 5",				EAN_8_5,
	"IATA",						IATA,
	"ISBT-128",					ISBT_128,
	"ISBT-128 concatenated",	ISBT_128_concatenated,
	"ITF",						ITF,
	"Macro PDF",				Macro_PDF,
	"MSI",						MSI,
	"PDF-417",					PDF_417,
	"Postbar (Canada)",			Postbar_Canada,
	"Postnet (US)",				Postnet_US,
	"Postal (Australia)",		Postal_Australia,
	"Postal (Japan)",			Postal_Japan,
	"Postal (UK)",				Postal_UK,
	"QR-code",					QR_code,
	"RSS-limited",				RSS_limited,
	"RSS-14",					RSS_14,
	"RSS-expanded",				RSS_Expanded,
	"Signature",				Signature,
	"Tri-Optic",				Trioptic_Code_39,
	"UPC-A",					UPCA,
	"UPC-A + 2",				UPCA_2,
	"UPC-A + 5",				UPCA_5,
	"UPC-E",					UPCE,
	"UPC-E + 2",				UPCE_2,
	"UPC-E + 5",				UPCE_5,
	"UPC-E1",					UPCE1,
	"UPC-E1 + 2",				UPCE1_2,
	"UPC-E1 + 5",				UPCE1_5
};

#define MAX_BCD					(sizeof(bcodedefs) / sizeof(SBCD_DEFS))
#define BARCODE_BUFFER_SIZE		257
#define TIMESTAMP_STRING_SIZE	31


SInt32 GetPrintAndClearCodes(OPNInterface interface)
{
	//vars we'll need
	bool		RTC = false;
	bool		ASCII = false;
	UInt32		numberOfCodes = 0;
	UInt32		barcodeDataLength = 0;
	UInt8		barcodeLength = 0;
	UInt8		barcodeType = 0;
	SInt32		result = 0;
	char*		barcode = new char[BARCODE_BUFFER_SIZE];
	char*		timeStamp = new char[TIMESTAMP_STRING_SIZE];

	//read the barcodes from the device
	result = readData(interface, &numberOfCodes);

	if (result == STATUS_OK)
	{
		//get packet type (ASCII or binary; binary isn't supported by the OPN 2001)
		result = getASCIIMode(interface, &ASCII);
	
		if (result != STATUS_OK)
		{
			//spit out an error
			return result;
		}
	
		//get timestamp setting (on or off)
		result = getRTCMode(interface, &RTC);
	
		if (result != STATUS_OK)
		{
			//spit out an error
			return result;
		}
	
		//a header before the barcodes
		std::cout << "Read " << numberOfCodes << " Barcodes" << std::endl;
		std::cout << "-------------------------------------" << std::endl;
		
		//get each code and print it
		int i = 0;
		for (; i < numberOfCodes; i++)
		{
			barcodeDataLength = BARCODE_BUFFER_SIZE;
			result = getBarcode(interface, barcode, i, &barcodeDataLength);
			
			if (result != STATUS_OK)
			{
				//spit out an error
				return result;
			}
			
			//if the getting was successful, barcodeDataLength has the total length, including the length byte of the code
			barcodeLength = barcode[0];	//barcode data + time bytes, no length
			barcodeType = barcode[1];
			
			if ((barcodeLength > 0) && (ASCII))
			{
				//subtract the time bytes from the barcode
				if (RTC)
				{
					//4 time bytes and the barcode type
					barcodeLength -= 5;
				}
				else
				{
					//the barcode type
					barcodeLength--;
				}
				
				//format the barcode and print
				printf("%*.*s", barcodeLength, barcodeLength, (barcode + 2));
				
				//find the barcode type and print it out
				int index = 0;
				for (; index < MAX_BCD; index++)
				{
					if (bcodedefs[index].symbol == barcodeType)
					{
						printf(" %s", bcodedefs[index].symbologyName);
						break;
					}
				}
				
				if (index == MAX_BCD)
				{
					std::cout << " Unknown Symbology";
				}
				
				//print out the timestamp if available
				if (RTC)
				{
					result = timeStamp2String(interface, timeStamp, TIMESTAMP_STRING_SIZE, reinterpret_cast<UInt8*>(barcode + (barcodeDataLength - 4)));
					
					if (result != STATUS_OK)
					{
						//spit out an error
						return result;
					}
					
					std::cout << "\t" << timeStamp << std::endl;
				}
			}
			else
			{
				std::cout << "Can't print out barcode" << std::endl;
			}
			
			memset(barcode, 0, sizeof(barcode));
			memset(timeStamp, 0, sizeof(timeStamp));
		}
		
		//clear the data
		result = clearDataCmd(interface);
		
		if (result != STATUS_OK)
		{
			//spit out an error
			return result;
		}
		
		std::cout << std::endl;
		std::cout << std::endl;
	}
	
	delete[] timeStamp;
	delete[] barcode;
	
	return result;
}



SInt32 ParametersAndTime(OPNInterface interface)
{
	SInt32	result = 0;
	UInt8	param = 0x04;
	char	oldParamValue = 0;		//disable
	char	paramValue = 1;			//enable
	UInt32	paramValueLength = 1;
	
	//test set param command with "reject redundant barcode" parameter
	result = setParamCmd(interface, param, &paramValue, paramValueLength);
	if (result != STATUS_OK)
	{
		std::cout << "Error Setting Param: " << result << std::endl;
		return result;
	}
	
	//make sure it was set
	char*	verifyParamValue = new char[4];
	UInt32	verifyParamValueLength = 4;
	result = getParamCmd(interface, param, verifyParamValue, &verifyParamValueLength);
	if (result != STATUS_OK)
	{
		std::cout << "Error Getting Param: " << result << std::endl;
		return result;
	}
	else
	{
		if (verifyParamValue[verifyParamValueLength - 1] == paramValue)
		{
			printf("Set Param Command Passed - Param: 0x%02x  Param Value: 0x%02x  Param Value Length: %d\n", param, 
				   (verifyParamValue[verifyParamValueLength - 1]), (verifyParamValueLength - 1));
		}
		else
		{
			std::cout << "Set Param Command Failed" << std::endl;
		}
	}
	
	//test set defaults command
	result = setDefaultsCmd(interface);
	if (result != STATUS_OK)
	{
		std::cout << "Error Setting Defaults: " << result << std::endl;
		return result;
	}
	
	//make sure it worked
	memset(verifyParamValue, 0, 4);
	verifyParamValueLength = 4;
	result = getParamCmd(interface, param, verifyParamValue, &verifyParamValueLength);
	if (result != STATUS_OK)
	{
		std::cout << "Error Getting Param: " << result << std::endl;
		return result;
	}
	else
	{
		if (verifyParamValue[verifyParamValueLength - 1] == oldParamValue)
		{
			printf("Set Defaults Command Passed - Param: 0x%02x  Param Value: 0x%02x  Param Value Length: %d\n", param, 
				   (verifyParamValue[verifyParamValueLength - 1]), (verifyParamValueLength - 1));
		}
		else
		{
			std::cout << "Set Defaults Command Failed" << std::endl;
		}
	}
	
	delete[] verifyParamValue;
	
	
	
	//time
	
	UInt8*	buf = new UInt8[6];
	UInt32	bufferLength = 6;
	time_t	timer;
	tm*		sysTime;
	
	//test the get time command
    result = getTimeCmd(interface, buf, &bufferLength);
	if (result != STATUS_OK)
	{
		std::cout << "Error Getting Time: " << result << std::endl;
		return result;
	}
	else
	{
		printf("Time: %02d:%02d:%02d  %02d/%02d/%02d\n", buf[2], buf[1], buf[0], buf[4], buf[3], buf[5]);
	}
	
	memset(buf, 0, 6);
	
	//test the set time command
	timer = time(NULL);
	sysTime = localtime(&timer);
	
	buf[0] = static_cast<UInt8>(sysTime->tm_sec);			// seconds
	buf[1] = static_cast<UInt8>(sysTime->tm_min);			// minutes
	buf[2] = static_cast<UInt8>(sysTime->tm_hour);			// hours
	
	buf[3] = static_cast<UInt8>(sysTime->tm_mday);			// day
	buf[4] = static_cast<UInt8>((sysTime->tm_mon) + 1);		// Month
	buf[5] = static_cast<UInt8>(sysTime->tm_year % 100);    // year
	
	result = setTimeCmd(interface, buf, bufferLength);
	if (result != STATUS_OK)
	{
		std::cout << "Error Setting Time: " << result << std::endl;
		return result;
	}
	else
	{
		std::cout << "Time Set" << std::endl;
	}
	
	delete[] buf;
	
	return result;
}



pascal OSStatus OPNEventHandler(EventHandlerCallRef nextHandler, EventRef theEvent, void* userData)
{
	//get the class and kind of the event
	UInt32 theEventClass = GetEventClass(theEvent);
	UInt32 theEventKind = GetEventKind(theEvent);
	
	//get our OPN interface
	OPNInterface interface = (OPNInterface) userData;
	
	//if it's an event from the OPN driver
	if (theEventClass == getEventClass())
	{
		//if it's an insertion event
		if (theEventKind == getEventKindInsert())
		{
			if (ParametersAndTime(interface) != STATUS_OK)
			{
				std::cout << "Error testing parameter manipulation functions and time get/set commands." << std::endl;
			}
			
			return noErr;
		}
		else if (theEventKind == getEventKindRemove())
		{
			
			return noErr;
		}
		else if (theEventKind == getEventKindDataAvailable())
		{
			if (GetPrintAndClearCodes(interface) != STATUS_OK)
			{
				std::cout << "Error getting, printing, and clearing barcodes from device." << std::endl;
			}
			
			return noErr;
		}
	}
	
	return eventNotHandledErr;
}


int main (int argc, char * const argv[]) {
	
	//install event handlers
	EventHandlerUPP OPNHandlerUPP;
	EventTypeSpec	OPNEventType[3];
	
	OPNEventType[0].eventClass = getEventClass();
	OPNEventType[0].eventKind = getEventKindInsert();

	OPNEventType[1].eventClass = getEventClass();
	OPNEventType[1].eventKind = getEventKindRemove();
	
	//this event is optional; a clear use would be to request a download of barcode data,
	//because the event is raised when there is data to read
	OPNEventType[2].eventClass = getEventClass();
	OPNEventType[2].eventKind = getEventKindDataAvailable();
	
	OPNHandlerUPP = NewEventHandlerUPP(OPNEventHandler);
	
	//create the OPN interface, so we can pass it to our event handler, which lets us interact with the object created here
	//without it having to be a global
	OPNInterface interface = CreateOPNInterface();
	
	OSStatus status;
	if ((status = InstallApplicationEventHandler(OPNHandlerUPP, 3, OPNEventType, interface, NULL)) != noErr)
	{
		//we have a problem here, but what do we do about it?  exception throwing?
	}

	
	//get and print the library version
	char* version = new char[LIBRARY_VERSION_SIZE];
	getLibraryVersion(interface, version, LIBRARY_VERSION_SIZE);
	
	std::cout << "Library Version: " << version << std::endl;
	
	delete[] version;
	
	//enable polling
	enablePolling(interface);
	
	//this is necessary to put the application in a state that makes it ready to receive and respond to events
	//once the point of execution passes this function (this function returns), the application exits; it's essentially
	//the main processing thread, getting events and sending them to the right places
	RunApplicationEventLoop();
	
	//delete event handler UPPs
	DisposeEventHandlerUPP(OPNHandlerUPP);
	
	//disable polling
	disablePolling(interface);
	
	//destroy the OPN interface
	DestroyOPNInterface(interface);

	
	return 0;	
}

/***********************************************************************************************
*	COOKIE CODE FOR EVENT HANDLER
***********************************************************************************************/

/* NOTE: These code fragments can be placed in the if-else:if-else:if block of the event handler,
 *		depending on what you want done when a certain event is raised.
 */




/*	This fragment sends the interrogate command and gets the Software Version and Device ID

char* SV = new char[SOFTWARE_VERSION_SIZE];
char* DevID = new char[DEVICE_ID_SIZE];
SInt32 result = 0;

//call interrogate
if ((result = interrogateCmd(interface)) != 0)
{
	std::cout << "Error: " << result << std::endl;
}

if (result == 0)
{
	//get the software version
	if ((result = getSoftwareVersion(interface, SV, SOFTWARE_VERSION_SIZE)) != 0)
	{
		std::cout << "Error: " << result << std::endl;
	}
	else
	{
		int i = 0;
		for (; i < SOFTWARE_VERSION_SIZE; i++)
		{
			//TODO: have to put some numeric manipulator in first...
			std::cout << std::hex << (UInt8) SV[i] << " ";
		}
		std::cout << std::endl;
	}
				
	//get the software version
	if ((result = getDeviceID(interface, DevID, DEVICE_ID_SIZE)) != 0)
	{
		std::cout << "Error: " << result << std::endl;
	}
	else
	{
		int i = 0;
		for (; i < DEVICE_ID_SIZE; i++)
		{
			//TODO: have to put some numeric manipulator in first...
			std::cout << std::hex << (UInt8) DevID[i] << " ";
		}
		std::cout << std::endl;
	}
}

delete[] SV;
delete[] DevID;

*/
